/*
Copyright 2004  Sukru Cinar

This file is part of rxsheyi.

rxsheyi is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

rxsheyi is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with rxsheyi; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <stdarg.h>

#define CLL_INSERT_BEFORE(x,y) do { \
                (y)->prev= (x)->prev; \
                (y)->next= (x); \
                (x)->prev->next= (y); \
                (x)->prev= (y); \
} while(0)

#define CLL_REMOVE(__ll_x) do { \
        (__ll_x)->prev->next= (__ll_x)->next; \
        (__ll_x)->next->prev= (__ll_x)->prev; \
} while(0)


#define for_children(a,b) \
do {\
	t b; \
	b= a->first; \
	if (b) \
		do {

#define end_for(a,b) \
			b=b->next; \
		} while(b!=a->first);\
} while(0)

typedef struct expression_s {
	int op;
	char *sdata;
	int idata;
	struct expression_s *prev,*next;
	struct expression_s *first;
} expression_t, *t;

t newt() {
	expression_t *e= malloc(sizeof(*e));
	e->prev= e->next= e;
	e->first= 0;
	return e;
}

char* newklass() {
	char *u= malloc(256);
	memset(u,0,256);
	return u;
}
	

t deep_copy(t x) {
	t u;
	u= newt();
	u->op= x->op;
	u->sdata= x->sdata;
	u->idata= x->idata;
	
	for_children(x,c)
		t cc= deep_copy(c);
		if (!u->first)
			u->first= cc;
		else
			CLL_INSERT_BEFORE(u->first,cc);
	end_for(x,c);
	return u;
}

t addchild(t p,t c) {
	c= deep_copy(c);
	if (!p->first)
		p->first= c;
	else
		CLL_INSERT_BEFORE(p->first,c);
	return p;
}

t setop(t p,int x) {
	p->op= x;
	return p;
}


t Star(t e) { return addchild( setop(newt(), '*'),e); }
t Plus(t e) { return addchild( setop(newt(), '+'),e); }
t Opt(t e) { return addchild( setop(newt(), '?'),e); }
t Or(t e1,t e2) { return addchild( addchild( setop(newt(), '|'), e1 ), e2); }
t Cat(t e1,t e2) { return addchild( addchild( setop(newt(), '.'), e1 ), e2); }
t Str(char *s) {
	t x= setop(newt(),'S');
	x->sdata= s;
	return x;
}
t Save(t e1,t e2) { return addchild(addchild(setop(newt(),'X'),e1),e2); }

t Number(int n) {
	t x= setop(newt(), 'N');
	x->idata= n;
	return x;
}

t OneOf(int x, ... ) {
	t o= setop(newt(),'O');
	o->sdata= newklass();
	va_list ap;
	va_start(ap,x);
	do {
		o->sdata[x]= 1;
		x= va_arg(ap,int);
	} while(x!=-1);
	va_end(ap);
	return o;
}

t Except(t e1,t e2) { return addchild( addchild( setop(newt(), 'E'), e1 ), e2); }

static int get_range_elem(t e) {
	if (e->op=='S') {
		if (strlen(e->sdata)!=1)
			exit(fprintf(stderr,"invalid range element %s\n",e->sdata));
		return e->sdata[0];
	} else if (e->op=='N')
		return e->idata;
	else
		exit(fprintf(stderr,"invalid argument to range functor type=%c\n",e->op));
}

t Range(t e1,t e2) {
	t o= setop(newt(),'O');
	int s,e,i;
	o->sdata= newklass();
	s= get_range_elem(e1);
	e= get_range_elem(e2);
	for(i=s;i<=e;i++)
		o->sdata[i]=1;
	return o;
}

int to_oneof(t c) {
	switch(c->op) {
	case 'S':
		if (strlen(c->sdata)==1) {
			char *z= newklass();
			z[(unsigned char) c->sdata[0]]= 1;
			c->op= 'O';
			c->sdata= z;
			break;
			return 0;
		}
	case '.':
	case '+':
	case '*':
	case '?':
	case 'X':
	case 'N':
		exit(fprintf(stderr,"invalid argument to Except()\n"));
		break;
	case '|':
	case 'E':
		return 1;
	case 'O':
		return 0;
	default:
		exit(fprintf(stderr,"unknown operator %c\n",c->op));
		break;
	}
	return 0;
}

static int combine_ranges(t x) {
	int ret=0;
	for_children(x,c) 
		ret+= combine_ranges(c);
	end_for(x,c);

	if (x->op!='|')
		return ret;

	for_children(x,c)
		if (c->op=='S' && strlen(c->sdata)==1) {
			char *z= newklass();
			z[(unsigned char) c->sdata[0]]= 1;
			c->op= 'O';
			c->sdata= z;
		}
	end_for(x,c);
	
	{ 
		t c=0;
		for_children(x,f)
			if (f->op=='O') {
				c=f;
				break;
			}
		end_for(x,f);
		if (c) {
			t n;
			t z;
			for(n=c->next;n!=x->first;n=z) {
				z= n->next;
				if (n->op=='O') {
					int i;
					for(i=0;i<256;i++)
						c->sdata[i] |= n->sdata[i];
					CLL_REMOVE(n);
					ret++;
				}
			}
		}
	}
	if (x->first->prev==x->first) {
		x->op= x->first->op;
		x->sdata= x->first->sdata;
		x->idata= x->first->idata;
		x->first= x->first->first;
	}
	return ret;
}

static void flatten(t x,int op) {
	for_children(x,c)
		flatten(c,op);
	end_for(x,c);

	if (x->op!=op)
		return ;

	while(1) {
		t f= 0;
		for_children(x,c)
			if (c->op==op) {
				f= c;
				break;
			}
		end_for(x,c);
		if (!f)
			return ;
		{
			t b= f->first;
			t e= f->first->prev;
			b->prev= f;
			f->next->prev= e;
			e->next= f->next;
			f->next= b;
			CLL_REMOVE(f);
			if (f==x->first)
				x->first= b;
		}
	}
}

void find_range(unsigned char *x,int w,int *Rb,int *Re) {
	int b,e;
	*Rb= *Re= 0;
	if (!x[w])
		return ;
	b= w;
	e= w;
	while(b>0 && x[b]) b--;
	if (b==0 || !x[b]) b++;
	while(e<256 && x[e]) e++;
	if (e==256 || !x[e])
		e--;
	*Rb= b;
	*Re= e;
}

int printchar(int x) {
	switch(x) {
		case '\n': return printf("\\n");
		case '\t': return printf("\\t");
		case '\r': return printf("\\r");
		case '\v': return printf("\\v");
		case '\b': return printf("\\b");
		case '\f': return printf("\\f");
		case '\\': return printf("\\\\");
		case '\"': return printf("\\\"");
	}
	if (!isprint(x))
		return printf("\\%03o",x);
	return printf("%c",x);
}

int printrxchar(int x) {
	switch(x) {
		case '\n': return printf("\\n");
		case '\t': return printf("\\t");
		case '\r': return printf("\\r");
		case '\v': return printf("\\v");
		case '\b': return printf("\\b");
		case '\f': return printf("\\f");
		case '\\': return printf("[\\\\]");
		case '\"': return printf("\\\"");
		case '(' :
		case ')' :
		case '+' :
		case '*' :
		case '?' :
		case '{' :
		case '}' :
		case '[' :
		case ']' :
		case '|' :
			return printf("\\\\%c",x);
	}
	if (!isprint(x))
		return printf("\\%03o",x);
	return printf("%c",x);
}

void prt_range(int b,int e) {
	if (!b)
		return ;
	if ((b<']' && e>=']') || (e-b)>3)  {
		printchar(b);
		printf("-");
		printchar(e);
	} else {
		while(b<=e)
			printchar(b++);
	}
}

void print_class(char *u) {
	int i;
	int db,de,re,rb;
	unsigned char x[256];
	memcpy(x, u , sizeof(x));
	printf("[");
	find_range(x, ']', &rb,&re);
	if (re==']') {
		printf("]");
		prt_range(rb,re-1);
	} else
		prt_range(rb,re);
	memset(x+rb, 0, re-rb+1);

	find_range(x, '-', &db,&de);
	if (db!=de)
		prt_range(db,de);
	memset(x+db, 0, de-db+1);

	for(i=1;i<256;i++)
		if (x[i]) {
			int b,e;
			find_range(x, i, &b,&e);
			prt_range(b,e);
			memset(x+b,0,e-b+1);
			i= e+1;
		}
	if (db==de && db) 
		printf("-");
	printf("]");
}

void print_str(expression_t *x) {
	switch(x->op) {
		case '|':
			for_children(x, c)
				if (c!=x->first)
					printf("|");
				print_str(c);
			end_for(x,c);
			break;
		case '.':
			for_children(x,c)
				if (c->op=='|')
					printf("(");
				print_str(c);
				if (c->op=='|')
					printf(")");
			end_for(x,c);
			break;
		case '?':
		case '+':
		case '*':
			if (x->first->op=='|' || x->first->op=='.')
				printf("(");
			print_str(x->first);
			if (x->first->op=='|' || x->first->op=='.')
				printf(")");
			printf("%c",x->op);
			break;
		case 'S': {
				int i;
				for(i=0;x->sdata[i];i++)
					printrxchar(x->sdata[i]);
			}
			break;
		case 'X':
			printf("(");
			print_str(x->first->next);
			printf(")");
			break;
		case 'O':
			print_class(x->sdata);
			break;
		default:
			exit(fprintf(stderr,"unknown operator %c\n",x->op));
	}
}

static void generate_saves(expression_t *x,char *prefix,int *p) {
	switch(x->op) {
		case '|':
			for_children(x, c)
				generate_saves(c,prefix,p);
			end_for(x,c);
			break;
		case '.':
			for_children(x,c)
				if (c->op=='|')
					(*p)++;
				generate_saves(c,prefix,p);
			end_for(x,c);
			break;
		case '?':
		case '+':
		case '*':
			if (x->first->op=='|' || x->first->op=='.')
				(*p)++;
			generate_saves(x->first,prefix,p);
			break;
			break;
		case 'X':
			(*p)++;
			printf("#define %s%s %d\n",prefix, x->first->sdata, *p);
			generate_saves(x->first->next,prefix,p);
			break;
		case 'O':
		case 'S':
			break;
		default:
			exit(fprintf(stderr,"unknown operator %c\n",x->op));
	}
}

static int calc_except(t x) {
	int ret= 0;
	for_children(x,c)
		ret+= calc_except(c);
	end_for(x,c);
	if (x->op=='E') {
		/* ranges have been combined and exceptions have been done (below this level) */
		t u= x->first;
		t v= x->first->next;
		int i;
		if (to_oneof(u) || to_oneof(v))
			return ret ;
		x->op='O';
		x->sdata= newklass();
		for(i=0;i<256;i++)
			x->sdata[i]= u->sdata[i] & ~v->sdata[i];
		x->first= 0;
		ret++;
	}
	return ret;
}

int dbg_print(t x) {
	static int tab= 3;
	int i;
	for(i=0;i<tab;i++)
			printf(" ");
	printf("%c ",x->op);
	switch(x->op) {
		case 'S': {
			int i;
			for(i=0;x->sdata[i];i++)
				printchar(x->sdata[i]);
			break;
		}
		case 'O':
			print_class(x->sdata);
			break;
	}
	printf("\n");
	tab+=3;
	for_children(x,c)
		dbg_print(c);
	end_for(x,c);
	tab-=3;
	return 0;
}

void _Generate(t x, char *strname,char *saveprefix) {
/*	printf("before flattening . \n"); dbg_print(x);*/
	flatten(x,'.');
/*	printf("before flattening | \n"); dbg_print(x);*/
	flatten(x, '|');
/*	printf("after flattening | \n"); dbg_print(x);*/
	while(combine_ranges(x) || calc_except(x)) {
/*		printf("after one step\n");
		dbg_print(x);*/
	}
		
	printf("#define %s \"",strname);
	print_str(x);
	printf("\"\n");
	{
		int s=0;
		generate_saves(x, saveprefix, &s);
		printf("#define %sCOUNT %d\n",saveprefix,s);
	}
}
