/*
Copyright 2004  Sukru Cinar

This file is part of rxsheyi.

rxsheyi is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

rxsheyi is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with rxsheyi; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
%{
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <sys/stat.h>
#include "option.h"

char *prgname;

typedef unsigned char uint8;

typedef struct expression_s {
	struct expression_s *prev,*next;
	int opcode;
	void *data; /* for stuff that read from external world */
	struct expression_s *first;
} expression_t;

#define for_children(a,b) \
do {\
	expression_t *b; \
	b= a->first; \
	if (b) \
		do {

#define end_for(a,b) \
			b=b->next; \
		} while(b!=a->first);\
} while(0)


typedef struct {
        int len,size;
        uint8 *data;
} cbuffer_t;

cbuffer_t *rest;

%}
%union {
	char *token;
	char ch;
	int dummy;
	cbuffer_t *cbuffer;
	expression_t *expression;
}
%token <token> IDENTIFIER NUMBER STRING_LITERAL
%token <ch> CHAR XCHAR 
%token <dummy> ONEOF MODMOD
%type  <cbuffer> rest charlist
%type <dummy> everything file element rule directive
%type <token> name
%type <expression> definition subdefinition factor alternative arglist arg funcexpr
%{

#define CLL_INSERT_BEFORE(x,y) do { \
                (y)->prev= (x)->prev; \
                (y)->next= (x); \
                (x)->prev->next= (y); \
                (x)->prev= (y); \
} while(0)


#define CLL_REMOVE(__ll_x) do { \
        (__ll_x)->prev->next= (__ll_x)->next; \
        (__ll_x)->next->prev= (__ll_x)->prev; \
} while(0)


static expression_t *new_expression() {
	expression_t *e= malloc(sizeof(*e));
	e->prev= e->next= e;
	e->first= 0;
	return e;
}

static expression_t *make_leaf(int code, void *data) {
	expression_t *e= new_expression();
	e->opcode= code;
	e->data= data;
	return e;
}

static expression_t* make_operator(int code, expression_t *sub, ... ) {
	expression_t *e= new_expression();
	e->opcode= code;
	if (sub) {
		va_list ap;
		e->first= sub;
		va_start(ap,sub);
		while((sub=va_arg(ap,expression_t*))) 
			CLL_INSERT_BEFORE(e->first,sub);
		va_end(ap);
	}
	return e;
}

static expression_t *expression_copy(expression_t *e) {
	expression_t *n= new_expression();
	n->opcode= e->opcode;
	n->data= e->data;
	if (e->first) {
		expression_t *c;
		c= e->first;
		do {
			expression_t *tc= expression_copy(c);
			if (n->first)
				CLL_INSERT_BEFORE(n->first,tc);
			else
				n->first= tc;
			c= c->next;
		} while(c!=e->first);
	}
	return n;
}

void cbuffer_init(cbuffer_t *cb) {
        cb->data= 0;
        cb->size= 0;
        cb->len= 0;
}

int cbuffer_append(cbuffer_t *cb,const uint8 *b,int len) {
        if (!len)
                return 0;
        if (cb->len+len>cb->size) {
                int ns= 1024;
                while(ns< (cb->len+len))
                        ns= ns+ ns/2;
                void *np= realloc(cb->data, ns);
                if (!np)
                        return 1;
                cb->data= np;
                cb->size= ns;
        }
        memcpy(cb->data+cb->len, b, len);
        cb->len+= len;
        return 0;
}

int cbuffer_appendc(cbuffer_t *cb,int x) {
	uint8 a= x;
	return cbuffer_append(cb, &a,1);
}

void cbuffer_uninit(cbuffer_t *cb) {
        if (cb->data)
                free(cb->data);
        cbuffer_init(cb);
}

typedef struct rule_s {
	struct rule_s *prev,*next;
	char *name;
	expression_t *definition;
} rule_t;

static rule_t head= { &head, &head };

static rule_t *find_rule(char *name) {
	rule_t *r;
	for (r=head.next;r!=&head;r=r->next)
		if (!strcmp(name,r->name))
			return r;
	return 0;
}

static void add_rule(char *name,expression_t *def) {
	rule_t *r= find_rule(name);
	if (r) {
		r->definition= make_operator('|' , r->definition, def, 0 );
	} else {
		r= malloc(sizeof(*r));
		r->name= name;
		r->definition= def;
		CLL_INSERT_BEFORE(&head,r);
	}
}

static void do_directive(char *name, expression_t *args);


int yylex();
static void yyerror(char *msg);
%}
%%
everything: file MODMOD rest { rest= $3; $$=$1; }
everything: file ;
file : file element ;
file : element;
element : rule { $$= 0; };
element : directive { $$=0; };

rule: name ':' definition ';' {
	add_rule($1,$3);
	$$= 0;
};

name : IDENTIFIER;
definition : subdefinition ;
definition : ONEOF charlist { $$=make_leaf('O', $2); } ;

charlist : charlist CHAR {
	cbuffer_appendc($1, $2);
	$$= $1;
};
charlist : CHAR {
	cbuffer_t *b= malloc(sizeof(*b));
	cbuffer_init(b);
	cbuffer_appendc(b, $1);
	$$=b;
};

subdefinition : subdefinition '|' alternative { $$= make_operator( '|' , $1, $3, 0 ); } ;
subdefinition : alternative;

alternative : alternative factor { $$= make_operator('.', $1, $2, 0 ); };
alternative : factor ;

factor : factor '*' { $$= make_operator( '*', $1, 0); };
factor : factor '?' { $$= make_operator( '?', $1, 0); };
factor : factor '+' { $$= make_operator( '+', $1, 0); };
factor : funcexpr ;
factor : '(' subdefinition ')'{ $$=$2; };
factor : IDENTIFIER { $$=make_leaf( 'I', $1); };
factor : STRING_LITERAL { $$= make_leaf( 'S', $1); };

funcexpr : IDENTIFIER '['  arglist ']' { $$=make_operator('[', make_leaf( 'I', $1), $3,0); };
arglist : arglist  ',' arg { $$= make_operator(',' , $1, $3,0); } ;
arglist : arg ;

arg: definition;
arg: NUMBER { $$= make_leaf( 'N', $1 ); };

directive : '.' name '[' arglist ']' ';' {
	do_directive($2,$4);
};

rest: rest XCHAR {
	cbuffer_appendc($1,$2);
	$$= $1;
};

rest: XCHAR {
	cbuffer_t *b= malloc(sizeof(cbuffer_t));
	cbuffer_init(b);
	cbuffer_appendc(b, $1);
	$$= b;
};

%%

#include "lex.yy.c"

static void yyerror(char *msg) {
	fprintf(stderr,"near line %d: %s\n",yylineno,msg);
	exit(1);
}

static FILE* output_file;

static int oprintf(char *fmt, ... ) {
	va_list ap;
	int ret;
	va_start(ap,fmt);
	ret= vfprintf(output_file, fmt, ap);
	va_end(ap);
	return ret;
}

static char *encode_char(int x) {
	static char buf[10];
	if (x<32 || x>126)
		sprintf(buf,"\\%03o",x);
	else if (x!='\\')
		sprintf(buf,"%c",x);
	else
		sprintf(buf,"\\\\");
	return buf;
}

static void print_code( expression_t *expr );

static void print_op(expression_t *e,char *fn,int trailing_zero) {
	expression_t *c;
	oprintf("%s(",fn);
	c= e->first;
	if (c)
		do {
			if (c!=e->first)
				oprintf(",");
			print_code(c);
			c= c->next;
		} while(c!=e->first);
	if (trailing_zero)
		oprintf(",0");
	oprintf(")");
}

static int builtin_operator(char *n) {
	static char * bi[]= {
		"Except", "Range", "Save",
		0
	};
	int i;
	for(i=0;bi[i];i++)
		if (!strcmp(n,bi[i]))
			return 1;
	return 0;
}

static void print_code( expression_t *expr ) {
	switch( expr->opcode ) {
		case '.': print_op(expr,"Cat",0); break;
		case '|': print_op(expr, "Or" ,0); break;
		case '?': print_op(expr, "Opt",0); break;
		case '+': print_op(expr, "Plus",0); break;
		case '*': print_op(expr, "Star",0); break;
		case '[': {
				expression_t *i= expr->first;
				expr->first= expr->first->next;
				CLL_REMOVE(i);
				print_op(expr, i->data, !builtin_operator(i->data));
				CLL_INSERT_BEFORE(expr->first,i);
				expr->first= i;
			}
			break;
		case ',':
			print_code(expr->first);
			oprintf(",");
			print_code(expr->first->next);
			break;
		case 'S':
			oprintf("Str(%s)", (char*) expr->data);
			break;
		case 'I':
			oprintf("__generate__%s()", (char*) expr->data);
			break;
		case 'N':
			oprintf("Number(%s)",(char*)  expr->data);
			break;
		case 'O': {
			int i;
			cbuffer_t *cb;
			oprintf("OneOf(");
			cb=expr->data;
			for(i=0;i<cb->len;i++)
				oprintf("'%s', ", encode_char(cb->data[i]));
			oprintf("-1)");
			break;
		}
	}
}

static void print_rule(rule_t *r) {
	oprintf("t __generate__%s() {\n\t return ",r->name);
	print_code(r->definition);
	oprintf(";\n}\n");
}

static void print_all_rules() {
	rule_t *r;
	for(r=head.next;r!=&head;r=r->next)
		print_rule(r);
}


static expression_t po[]= {
	{ 0, 0, 0, 0 },
	{ 0, 0, 0, "Star", 0 },
	{ 0, 0, 0, "Plus", 0 },
	{ 0, 0, 0, "Opt", 0 },
	{ 0, 0, 0, "Or", 0 },
	{ 0, 0, 0, "Cat", 0 },
	{ 0, 0, 0, "Save", 0},
	{ 0, 0, 0, "Range", 0 },
	{ 0, 0, 0, "Except", 0 },
};

void print_operator_prototypes(expression_t* x) {
	for_children(x,c)
		print_operator_prototypes(c);
	end_for(x,c);
	if (x->opcode=='[') {
		char *y= x->first->data;
		if (!po[0].prev) {
			int i;
			int sz= sizeof(po)/sizeof(po[0]);
			for(i=0;i<sz;i++) {
				po[i].prev= po+(i+sz-1)%sz;
				po[i].next= po+(i+1)%sz;
			}
		} 
		for(x=po+1;x!=po;x=x->next)
			if (!strcmp(y,x->data))
				return ;
		x= new_expression();
		x->data= y;
		CLL_INSERT_BEFORE(po, x);
		oprintf( "t %s(t x, ... );\n",y); 
	}
}

static void print_prototypes() {
	rule_t *r;
	for(r=head.next;r!=&head;r=r->next)
		oprintf("t __generate__%s();\n",r->name);
	for(r=head.next;r!=&head;r=r->next)
		print_operator_prototypes(r->definition);
}


static expression_t toplevels= { &toplevels, &toplevels };

static expression_t *getarg(expression_t *t,int x,int *y) {
	if (t->opcode!=',') {
		if (*y==x)
			return t;
		(*y)++;
		return NULL;
	} else {
		expression_t *left= getarg(t->first, x, y);
		if (left)
			return left;
		return getarg(t->first->next,x,y);
	}
}

static expression_t *arg(expression_t *args,int i) {
	int x=0;
	return getarg(args,i,&x);
}

static void do_generate(expression_t *args) {
	expression_t *a[3];
	int i;
	for(i=0;i<3;i++) a[i]= arg(args,i);
	if (a[0]->opcode!='I' || a[1]->opcode!='S' || a[2]->opcode!='S')
		exit(fprintf(stderr,"error while generating expression;\nsyntax: .generate[ symbol, \"strname\", \"savenameprefix\"\n"));
	oprintf("_Generate( __generate__%s(), %s, %s);\n", (char*) a[0]->data, (char*) a[1]->data, (char*) a[2]->data);
}

static void do_directive(char *name, expression_t *args) {
	if (!strcmp(name,"generate"))
		CLL_INSERT_BEFORE(&toplevels,args);
}


static FILE *input_file;
static char* output_file_name;
static char* source_output_file_name;
static char* temp_directory;
static char *compiler="cc";
static int compargs=0;
static char *compiler_args[20]= {
	"-g", "-o", "%o", "%i", 0 };
static char *exe_file_name;
static int nocompile=0,noexec= 0;

static char *make_temp_file() {
	int i;
	char host[200];
	int r;
	struct stat sbuf;
	gethostname(host, sizeof(host));
	host[sizeof(host)-1]= 0;
	if (!temp_directory) {
		temp_directory=getenv("TMP_DIR");
		if (!temp_directory)
			temp_directory=getenv("HOME");
		if (!temp_directory)
			exit(fprintf(stderr,"%s: either give a temp directory option, or set one of HOME or TMP_DIR\n",prgname));
	}
	i=0;
	while ((r=stat(temp_directory,&sbuf)) && errno==EINTR)
		;
	if (r)
		exit(fprintf(stderr,"%s: can not stat temp directory %s: %s\n",prgname, temp_directory, strerror(errno)));
	while(1) {
		char buf[2000];
		snprintf(buf,sizeof(buf),"%s/rxgen%s-%d-%d.c",temp_directory, host, getpid(), i);
		while ((r=stat(buf,&sbuf)) && errno==EINTR)
			;
		if (r && errno==ENOENT)
			return strdup(buf);
		if (!r)
			i++;
		else
			exit(fprintf(stderr,"%s: while finding a temporary file name in %s: %s\ntry giving one by hand and make sure its name ends with .c\n",prgname,temp_directory,strerror(errno)));
	}
	return 0;
}

int parse_options(int argc,char **argv) {
	OPTIONS_BEGIN
	OPTION( "-o", ARG_STR, "Print the output to the given file instead of stdout" ) {
		output_file_name= arg;
	}
	OPTION( "-s", ARG_STR, "Print the intermediate code to given file") {
		source_output_file_name= arg;
	}
	OPTION( "-x", ARG_STR, "Compile the intermediate program into this file") {
		exe_file_name= arg;
	}
	OPTION( "-t", ARG_STR, "Set the temporary directory for intermediate files") {
		temp_directory= arg;
	}
	OPTION( "-cc", ARG_STR, "Set the compiler") { compiler= arg; }
	OPTION( "-cf", ARG_STR, "Append an argument or flag to the compiler invocation.\n"
				"\t\t%i is the input file %o is the output file.\n"
				"\t\tYou should give this option once for each compiler option") {
		if (compargs==19)
			exit(fprintf(stderr,"%s: too many compiler flags\n",argv[0]));
		compiler_args[compargs++]= arg;
	}
	OPTION( "-nc", ARG_NONE, "Compilation is not done if you give this flag. Only the source is generated") {
		nocompile= 1;
	}
	OPTION( "-nx", ARG_NONE, "Execution is not done if you give this flag. Only the executable is generated") {
		noexec= 1;
	}
	ARG {
		if (input_file)
			exit(fprintf(stderr,"%s: did not understand second argument %s\n",argv[0],arg));
		input_file= fopen(arg,"r");
		if (!input_file)
			exit(fprintf(stderr,"%s: can not open file %s for reading: %s\n", argv[0], arg, strerror(errno)));
	}
	OPTIONS_END
}

int main(int argc,char **argv) {
	prgname= argv[0];
	parse_options(argc,argv);
	if (!input_file)
		input_file= stdin;
	yyin= input_file;
	yyparse();

	if (!source_output_file_name)
		source_output_file_name= make_temp_file();
	output_file= fopen(source_output_file_name, "w");
	if (!output_file)
		exit(fprintf(stderr,"%s: can not open %s for writing: %s\n",argv[0], source_output_file_name, strerror(errno)));

	{
		static const char libdotc[]= {
#include "lib.c.hex"
		0
		};
		oprintf("%s",libdotc);
	}
	print_prototypes();
	print_all_rules();

	if (rest) {
		fwrite(rest->data,1,rest->len,output_file);
	}

	oprintf("int main() {\n");
	{
		expression_t *e;
		for(e=toplevels.next;e!=&toplevels;e=e->next)
			do_generate(e);
	}
	oprintf("	return 0;\n}\n");
	fclose(input_file);
	fclose(output_file);

	if (nocompile)
		return 0;
	{
		int len=0;
		char *buf;
		char *args[22];
		int i;
		if (!exe_file_name)
			exe_file_name= make_temp_file();
		if (compargs==0)
			while(compiler_args[compargs])
				compargs++;
		for(i=0;i<compargs;i++) {
			if (!strcmp(compiler_args[i],"%o"))
				args[i+1]= exe_file_name;
			else if (!strcmp(compiler_args[i],"%i"))
				args[i+1]= source_output_file_name;
			else
				args[i+1]= compiler_args[i];
			len+= strlen(args[i+1])+3;
		}
		args[0]= compiler;
		len+= strlen(compiler)+1;
		buf= malloc(len+1);
		buf[0]= 0;
		for(i=0;i<compargs+1;i++) {
			strcat(buf," ");
			if (i) strcat(buf,"'");
			strcat(buf,args[i]);
			if (i) strcat(buf,"'");
		}
		fprintf(stderr,"%s: executing %s\n",argv[0],buf);
		if (system(buf))
			exit(fprintf(stderr,"%s: compilation failed\n",argv[0]));
	}
	if (noexec)
		return 0;
	if (system(exe_file_name))
		exit(fprintf(stderr,"%s: execution of generator %s failed\n",argv[0],exe_file_name));
	unlink(source_output_file_name);
	unlink(exe_file_name);
	return 0;
}
